@@ -1,9 +1,9 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2021 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2021-2024 -- leonerd@leonerd.org.uk
 
-package XS::Parse::Infix 0.21;
+package XS::Parse::Infix 0.44;
 
 use v5.14;
 use warnings;
@@ -11,6 +11,8 @@ use warnings;
 # No actual .xs file; the code is implemented in XS::Parse::Keyword
 require XS::Parse::Keyword;
 
+use Carp;
+
 =head1 NAME
 
 C<XS::Parse::Infix> - XS functions to assist in parsing infix operators
@@ -28,15 +30,32 @@ and subject to change. Later versions may break ABI compatibility, requiring
 changes or at least a rebuild of any module that depends on it.
 
 In addition, the places this functionality can be used are relatively small.
-No current release of perl actually supports custom infix operators, though I
-have a branch where I am currently experimenting with such support:
-
-L<https://github.com/leonerd/perl5/tree/infix-plugin>
+Support for custom infix operators as added in Perl development release
+C<v5.37.7>, and is therefore present in Perl C<v5.38.0>.
 
 In addition, the various C<XPK_INFIX_*> token types of L<XS::Parse::Keyword>
 support querying on this module, so some syntax provided by other modules may
 be able to make use of these new infix operators.
 
+=head2 Lexically-named vs. Globally-named Operators
+
+At version 0.43, the way this module operates has changed. Since this version,
+the preferred method of operation for infix operators is that the module that
+registers them creates them with a fully-qualified name (i.e. including the
+name of the package that implements them). Other code which wishes to import
+the operator can then provide a local name for it visible within a scope. This
+acts as an alias for the fully-qualified registered one. The imported name
+needs not be the same in various scopes - this allows imported operators to be
+renamed in specific scopes to avoid name clashes.
+
+Prior versions of this module only supported operators that have one global
+name that may or may not be enabled in lexical scopes. With this model, while
+the visiblity of an operator can be controlled per scope, if it is visible it
+must always have the same name and cannot be renamed to avoid a name clash.
+
+At some future version of this module the older method may be removed, once
+all of the existing CPAN modules have been updated to the new model.
+
 =cut
 
 =head1 CONSTANTS
@@ -49,11 +68,101 @@ This constant is true if built on a perl that supports the C<PL_infix_plugin>
 extension mechanism, meaning that custom infix operators registered with this
 module will actually be recognised by the perl parser.
 
-No actual production or development releases of perl yet support this feature,
-but see above for details of a branch which does.
+No actual production releases of perl yet support this feature, but see above
+for details of development versions which do.
 
 =cut
 
+=head1 PERL FUNCTIONS
+
+=head2 apply_infix
+
+   $pkg->apply_infix( $on, $args, @infix );
+
+I<Since version 0.43.>
+
+Intended to be called by the C<import> and C<unimport> methods of a module
+that implements infix operators. This sets up the various keys required in the
+lexical hints hash to create a lexical alias for the requested operator(s).
+
+I<$on> should be a true value for C<import> or a false value for C<unimport>.
+I<$args> should be an array reference to the arguments array, which will be
+modified to splice out recognised values, leaving behind anything else
+unrecognised. I<@infix> should be a list of names of operators implemented by
+the importing package, named by I<$pkg>.
+
+After invoking this method, the caller should inspect for any remaining values
+in the arguments array, either to handle them in some other way, or just
+complain about them.
+
+   package Some::Infix::Module;
+
+   sub import
+   {
+      my $pkg = shift;
+      $pkg->XS::Parse::Infix::apply_infix( 1, \@_, qw( operator names here ) );
+      die "Unrecognised symbols: @_" if @_;
+   }
+
+When this module is C<use>d, named operators are aliased from the
+fully-qualified names.
+
+   use Some::Infix::Module qw( operator );
+
+   my $z = $x operator $y;
+
+At this point, the name C<operator> becomes a lexical alias to
+C<Some::Infix::Module::operator>.
+
+Each imported operator name can optionally be followed by import options,
+given in a hash reference. Currently the only named option recognised is
+C<-as>, which allows the importing scope to provide a different name for the
+imported operator, perhaps to avoid name clashes, or for neatness.
+
+   use Some::Infix::Module operator => { -as => "myname" };
+
+   my $z = $x myname $y;  # invokes Some::Infix::Module::operator
+
+=cut
+
+# helper functions for infix operator modules
+sub apply_infix
+{
+   my $pkg = shift;
+   my ( $on, $args, @infix ) = @_;
+
+   my %infix = map { $_ => 1 } @infix;
+
+   for ( my $idx = 0; 1; $idx++ ) {
+      $idx < @$args or last;
+
+      my $name = $args->[$idx];
+      $infix{$name} or next;
+
+      splice @$args, $idx, 1, ();
+
+      my %opts;
+      %opts = %{ splice @$args, $idx, 1, () } if "HASH" eq ref $args->[$idx];
+
+      my $localname = $name;
+      $localname = delete $opts{-as} if exists $opts{-as};
+
+      croak "Unrecognised apply_infix options " . join( ", ", sort keys %opts )
+         if %opts;
+
+      XS::Parse::Infix::check_opname( $localname ) or
+         croak "Local name '$localname' for imported operator $name is invalid";
+
+      my $hintkey = "XS::Parse::Infix/$localname";
+      my $fqname  = "${pkg}::${name}";
+
+      $on ? $^H{$hintkey} = $fqname
+          : delete $^H{$hintkey};
+
+      redo;
+   }
+}
+
 =head1 XS FUNCTIONS
 
 =head2 boot_xs_parse_infix
@@ -68,6 +177,17 @@ requirement; e.g.
 
    boot_xs_parse_infix(0.14);
 
+=head2 parse_infix
+
+   bool parse_infix(enum XSParseInfixSelection select, struct XSParseInfixInfo **infop);
+
+I<Since version 0.27.>
+
+This function attempts to parse syntax for an infix operator from the current
+parser position. If it is successful, it fills in the variable pointed to by
+I<infop> with a pointer to the actual information structure and returns
+C<true>. If no suitable operator is found, returns C<false>.
+
 =head2 xs_parse_infix_new_op
 
    OP *xs_parse_infix_new_op(const struct XSParseInfixInfo *info, U32 flags,
@@ -79,7 +199,7 @@ perl's C<newBINOP> function.
 
 The C<info> structure pointer would be obtained from the C<infix> field of the
 result of invoking the various C<XPK_INFIX_*> token types from
-C<XS::Parse::Keyword>.
+C<XS::Parse::Keyword>, or by calling L</parse_infix> directly.
 
 =head2 register_xs_parse_infix
 
@@ -88,8 +208,8 @@ C<XS::Parse::Keyword>.
 
 This function installs a set of parsing hooks to be associated with the given
 operator name. This new operator will then be available via
-L<XS::Parse::Keyword> by the various C<XPK_INFIX_*> token types, or to core
-perl's C<PL_infix_plugin> if availble.
+L<XS::Parse::Keyword> by the various C<XPK_INFIX_*> token types,
+L</parse_infix>, or to core perl's C<PL_infix_plugin> if available.
 
 These tokens will all yield an info structure, with the following fields:
 
@@ -100,11 +220,17 @@ These tokens will all yield an info structure, with the following fields:
 
       struct XSParseInfixHooks *hooks;
       void                     *hookdata;
+
+      enum XSParseInfixClassification cls;  /* since version 0.28 */
    };
 
 If the operator name contains any non-ASCII characters they are presumed to be
 in UTF-8 encoding. This will matter for deparse purposes.
 
+If I<opname> contains a double-colon (C<::>) sequence, it is presumed to be a
+fully-qualified operator name in the new interface style. If not, it is
+presumed to be an older globally-named one.
+
 =cut
 
 =head1 PARSE HOOKS
@@ -113,7 +239,7 @@ The C<XSParseInfixHooks> structure provides the following fields which are
 used at various stages of parsing.
 
    struct XSParseInfixHooks {
-      U16 flags; /* currently ignored */
+      U16 flags;
       U8 lhs_flags;
       U8 rhs_flags;
       enum XSParseInfixClassification cls;
@@ -123,37 +249,76 @@ used at various stages of parsing.
       const char *permit_hintkey;
       bool (*permit)(pTHX_ void *hookdata);
 
-      OP *(*new_op)(pTHX_ U32 flags, OP *lhs, OP *rhs, void *hookdata);
+      OP *(*new_op)(pTHX_ U32 flags, OP *lhs, OP *rhs, SV **parsedata, void *hookdata);
       OP *(*ppaddr)(pTHX);
+
+      /* optional */
+      void (*parse)(pTHX_ U32 flags, SV **parsedata, void *hookdata);
    };
 
 =head2 Flags
 
-The C<flags> field is currently ignored. It is defined simply to reserve the
-space in case used in a later version. It should be set to zero.
-
-The C<rhs_flags> field gives details on how to parse and handle the right-hand
-side of the operator syntax. It should be set to one of the following constants:
+The C<flags> field gives details on how to handle the operator overall. It
+should be a bitmask of the following constants, or left as zero:
 
 =over 4
 
-=item XPI_OPERAND_TERM (0)
+=item XPI_FLAG_LISTASSOC
 
-Default. The operand is a term expression.
+I<Since version 0.40.>
+
+If set, the operator supports n-way list-associative syntax; written in the
+form
+
+   OPERAND op OPERAND op OPERAND op ...
+
+In this case, the custom operator will be a LISTOP rather than a BINOP, and
+every operand of the entire chain will be stored as a child op of it. The op
+function will need to know how many operands it is working on. There are two
+ways this may be indicated, depending on whether it was known at compile-time.
+
+If the number operands is known at compile-time, the C<OPf_STACKED> flag is
+not set, and the C<op_private> field indicates the number of operand
+expressions that were present.
+
+If the number is not known (for example, the operator is being used as the
+body of a generated wrapper function), the C<OPf_STACKED> flag is set. The
+number of arguments will be passed as the UV of an extra SV which is pushed
+last to the stack. The op function should pop this first to find out.
+
+It is typical to begin a list-associative op function with code such as:
+
+   int n = (PL_op->op_flags & OPf_STACKED) ? POPu : PL_op->op_private;
+
+If the operator is list-associative, then C<lhs_flags> and C<rhs_flags> must
+be equal.
+
+=back
+
+The C<lhs_flags> and C<rhs_flags> fields give details on how to handle the
+left- and right-hand side operands, respectively.
+
+It should be set to one of the following constants, or left as zero:
+
+=over 4
 
 =item XPI_OPERAND_TERM_LIST
 
-The operand is a term expression. It will be foced into list context,
-preserving the C<OP_PUSHMARK> at the beginning. This means that the ppfunc for
-this infix operator will have to C<POPMARK> to find that.
+The operand will be foced into list context, preserving the C<OP_PUSHMARK> at
+the beginning. This means that the ppfunc for this infix operator will have to
+C<POPMARK> to find that.
 
 =item XPI_OPERAND_LIST
 
-The operand is a list expression. It will be forced into list context, the
-same as above.
+The same as above.
 
 =back
 
+Older versions used to provide constants named C<XPI_OPERAND_ARITH> and
+C<XPI_OPERAND_TERM> but they related to an older version of the core perl
+branch. These names are now aliases for zero, and can be removed from new
+code.
+
 In addition the following extra bitflags are defined:
 
 =over 4
@@ -179,12 +344,6 @@ the values passed.
 
 =back
 
-The C<lhs_flags> field gives details on how to handle the left-hand side of
-the operator syntax. It takes similar values to C<rhs_flags>, except that it
-does not accept the C<XPI_OPERAND_LIST> value. Parsing always happens on just
-a term expression, though it may be placed into list context (which therefore
-still permits things like parenthesized lists, or array variables).
-
 =head2 The Selection Stage
 
 The C<cls> field gives a "classification" of the operator, suggesting what
@@ -210,6 +369,18 @@ Both the string and the function are optional. Either or both may be present.
 If neither is present then the keyword is always permitted - which is likely
 not what you wanted to do.
 
+=head2 The C<parse> Stage
+
+If the optional C<parse> hook function is present, it is called immediately
+after the parser has recognised the presence of the named operator itself but
+before it attempts to consume the right-hand side term. This hook function can
+attempt further parsing, in order to implement more complex syntax such as
+hyper-operators.
+
+When invoked, it is passed a pointer to an C<SV *>-typed storage variable. It
+is free to use this variable it desires to store a result, which will then
+later be made available to the C<new_op> function.
+
 =head2 The Op Generation Stage
 
 If the infix operator is going to be used, then one of the C<new_op> or the
@@ -217,8 +388,14 @@ C<ppaddr> fields explain how to create a new optree fragment.
 
 If C<new_op> is defined then it will be used, and is expected to return an
 optree fragment that consumes the LHS and RHS arguments to implement the
-semantics of the operator. If this is not present, then the C<ppaddr> will be
-used instead to construct a new BINOP of the C<OP_CUSTOM> type.
+semantics of the operator. If the optional C<parse> stage had been present
+earlier, the C<SV **> pointer passed here will point to the same storage that
+C<parse> had previously had access to, so it can retrieve the results.
+
+If C<new_op> is not present, then the C<ppaddr> will be used instead to
+construct a new BINOP or LISTOP of the C<OP_CUSTOM> type. If an earlier
+C<parse> stage had stored additional results into the C<SV *> variable these
+will be lost here.
 
 =head2 The Wrapper Function
 
@@ -311,6 +488,18 @@ C<OP_ANONLIST> type, which is used by anonymous array-ref construction:
 This optimisation is only performed if the operator declared it safe to do so,
 via the C<XPI_OPERAND_ONLY_LOOK> flag.
 
+If a function of the given name already exists at registration time it will be
+left undisturbed and no new wrapper will be created. This permits the same
+infix operator to have multiple spellings of its name; for example to allow
+both a real Unicode and a fallback ASCII transliteration of the same operator.
+The first registration will create the wrapper function; the subsequent one
+will skip it because it would otherwise be identical.
+
+Note that when generating an optree for a wrapper function call, the C<new_op>
+hook function will be invoked with a C<NULL> pointer for the C<SV *>-typed
+parse data storage, as there won't be an opporunity for the C<parse> hook to
+run in this case.
+
 =cut
 
 =head1 DEPARSE
@@ -393,13 +582,18 @@ sub B::Deparse::_deparse_infix_named
 {
    my ( $self, $opname, $op, $ctx ) = @_;
 
-   my $lhs = $op->first;
-   my $rhs = $op->last;
+   # Cope with list-associatives
+   my $kid = $op->first;
 
-   return join " ",
-      $self->deparse_binop_left( $op, $lhs, 14 ),
-      $opname,
-      $self->deparse_binop_right( $op, $rhs, 14 );
+   my @operands = ( $self->deparse_binop_left( $op, $kid, 14 ) );
+   $kid = $kid->sibling;
+
+   while( !B::Deparse::null( $kid ) ) {
+      push @operands, $self->deparse_binop_right( $op, $kid, 14 );
+      $kid = $kid->sibling;
+   }
+
+   return join " $opname ", @operands;
 }
 
 =head1 TODO
@@ -412,6 +606,12 @@ Have the entersub checker for list/list operators unwrap arrayref or
 anon-array argument forms (C<WRAPPERFUNC( \@lhs, \@rhs )> or
 C<WRAPPERFUNC( [LHS], [RHS] )>).
 
+=item *
+
+Further thoughts about how infix operators with C<parse> hooks will work with
+automatic deparse, and also how to integrate them with L<XS::Parse::Keyword>'s
+grammar piece.
+
 =back
 
 =cut
